import time
import unittest

from scapy.layers.dns import DNS, DNSRR
from scapy.layers.inet import IP, TCP, UDP, ICMP
from unittest.mock import patch
from scapy.layers.l2 import Ether, ARP

import duckdb as duckdb
from scapy.packet import Raw

import database_connection as db_conn

from packet_capturing import packet_handler, get_dicts, dict_clear, ip_rate_based_anomaly_detection, port_check, \
    ip_spoofing, \
    syn_fin, null_packet, tcp_reset_attack, arp_spoofing, dns_spoofing, checksum_verification, payload_pattern_matching, \
    content_length_mismatch, malformed_check, icmp_flood_detection, reserved_ips, protocol_fragmentation_check

connection = duckdb.connect('../../packets_data.duckdb')

class TestPacketCapturing(unittest.TestCase):

    # This method is called before every test
    def setUp(self):
        db_conn.initiate_database(connection)
        dict_clear()

    def test_baseCase(self):
        # Create a dummy packet
        # Ether addresses are placeholders
        packet = (IP(src = "100.84.6.141", dst = "192.168.1.1") /
                  TCP(sport=12345, dport=80, flags="S", seq=1000, options=[('Timestamp', (1234567890, 0))]) /
                  Ether(src="00:11:22:33:44:55", dst="ff:ff:ff:ff:ff:ff"))
        # Show detailed information about the package
        print(packet.show())

        # Mockito checks if print was called
        with patch('builtins.print') as mock_print:
            packet_handler(packet)
            mock_print.assert_any_call("WARNING: Suspicious packet detected.")

    def test_ip_spoofing(self):
        # Starts with 169.254, it's a suspicious packet
        packet = IP(src = "169.254.6.141", dst = "192.168.1.1")

        with patch('builtins.print') as mock_print:
            ip_spoofing(0, packet[IP].src, packet)
            mock_print.assert_any_call("Possible IP spoofing using private networks detected.")

        with patch('builtins.print') as mock_print:
            for reserved_ip in reserved_ips:
                ip_spoofing(0, IP(src = reserved_ip, dst = "192.168.1.1").src, packet)
                mock_print.assert_not_called()

    def test_syn_fin(self):
        # Has S and F flags, which is also known as SYN-FIN, which is a malicious packet signature

        packet = (IP(src = "100.84.6.141", dst = "192.168.1.1") /
                  TCP(dport = 80, flags = "SF") /
                  Ether(src="00:11:22:33:44:55", dst="ff:ff:ff:ff:ff:ff"))

        with patch('builtins.print') as mock_print:
            syn_fin(packet)
            mock_print.assert_any_call("Malicious packet detected: SYN-FIN combination.")

    def test_null_packet(self):
        packet = (IP(src = "100.84.6.141", dst = "192.168.1.1") /
                  TCP(dport = 80, flags = 0) /
                  Ether(src="00:11:22:33:44:55", dst="ff:ff:ff:ff:ff:ff"))

        with patch('builtins.print') as mock_print:
            null_packet(packet)
            mock_print.assert_any_call("Malicious null packet found.")

    def test_port_dest_check(self):
        packet = (IP(src="100.84.6.141", dst="192.168.1.1") /
                  TCP(dport = 0, sport = 200, flags="F") /
                  Ether(src="00:11:22:33:44:55", dst="ff:ff:ff:ff:ff:ff"))

        with patch('builtins.print') as mock_print:
            port_check(packet)
            mock_print.assert_any_call("Illegal packet with source or destination port 0.")

    def test_port_src_check(self):
        packet = (IP(src="100.84.6.141", dst="192.168.1.1") /
                  TCP(dport=80, sport=0, flags="F") /
                  Ether(src="00:11:22:33:44:55", dst="ff:ff:ff:ff:ff:ff"))

        with patch('builtins.print') as mock_print:
            packet_handler(packet)
            mock_print.assert_any_call("Illegal packet with source or destination port 0.")

    def test_ip_rate_based_detection(self):
        repetitions = 150
        packet = (IP(src="100.84.6.141", dst="192.168.1.1") /
                  TCP(dport=80, sport=123, flags="S") /
                  Ether(src="00:11:22:33:44:55", dst="ff:ff:ff:ff:ff:ff"))

        with patch('builtins.print') as mock_print:
            for i in range(repetitions):
                ip_rate_based_anomaly_detection(packet)
            time.sleep(1)
            ip_rate_based_anomaly_detection(packet)
            mock_print.assert_called()
        self.assertEqual({"100.84.6.141": 50}, get_dicts()[0], "Expected the packet's IP")

    def test_icmp_flood_detection(self):
        repetitions = 150
        packet = (IP(src="100.84.6.141", dst="192.168.1.1") /
                  ICMP(type=8, code=0) /
                  Ether(src="00:11:22:33:44:55", dst="ff:ff:ff:ff:ff:ff"))

        with patch('builtins.print') as mock_print:
            for i in range(repetitions):
                icmp_flood_detection(packet)
            self.assertEqual({"100.84.6.141": repetitions}, get_dicts()[2], f"Expected the packet's IP with {repetitions}.")
            time.sleep(1)
            icmp_flood_detection(packet)
            mock_print.assert_called()

    def test_tcp_reset_attack(self):
        packet = (IP(src="100.84.6.141", dst="192.168.1.1") /
                  TCP(dport=80, sport=0, flags="R") /
                  Ether(src="00:11:22:33:44:55", dst="ff:ff:ff:ff:ff:ff"))

        db_conn.add_address(connection, packet)
        with patch('builtins.print') as mock_print:
            tcp_reset_attack(packet)
            mock_print.assert_any_call(f"Possible TCP reset attack from {packet[IP].src}")
        self.assertTrue(db_conn.contains_address(connection=connection, packet=packet))

    def test_arp_spoofing(self):
        # hwsrc = Hardware Source Address
        # psrc = Protocol Source Address (iPv4 address)
        # op = ARP reply

        packet1 = (IP(src="100.84.6.141", dst="192.168.1.1") /
                  TCP(dport=80, sport=0, flags="R") /
                  Ether(src="00:11:22:33:44:55", dst="ff:ff:ff:ff:ff:ff") /
                  ARP(hwsrc= "00:11:22:33:44:55:66", psrc= "192.168.1.100", op = 2))

        packet2 = (IP(src="100.84.6.141", dst="192.168.1.1") /
                   TCP(dport=80, sport=0, flags="R") /
                   Ether(src="00:11:22:33:44:55", dst="ff:ff:ff:ff:ff:ff") /
                   ARP(hwsrc="00:11:22:33:44:55:00", psrc="192.168.1.100", op = 2))

        packet3 = (IP(src="100.84.6.141", dst="192.168.1.1") /
                   TCP(dport=80, sport=0, flags="R") /
                   Ether(src="00:11:22:33:44:55", dst="ff:ff:ff:ff:ff:ff") /
                   ARP(hwsrc="00:11:22:33:44:55:00", psrc="192.168.1.100", op=1))

        arp_spoofing(packet1)
        with patch('builtins.print') as mock_print:
            arp_spoofing(packet2)
            mock_print.assert_called()

        with patch('builtins.print') as mock_print:
            arp_spoofing(packet3)
            mock_print.assert_not_called()

    def test_dns_spoofing(self):
        packet = (Ether() /
                  IP(src="8.8.8.8", dst="192.168.1.1") /
                  UDP(dport=53) /
                  DNS(qr=1, ancount=1, an=DNSRR(rrname="example.com", rdata="192.168.1.4")))

        with patch('builtins.print') as mock_print:
            dns_spoofing(packet)
            mock_print.assert_called_with(f"Possible DNS spoofing detected: {packet[DNS].an.rrname} -> 192.168.1.4")

    def test_checksum_verification(self):
        packet = (IP(src="192.168.1.1", dst="192.168.1.2") /
                  TCP(dport=80, sport=12345) /
                  Ether())
        packet.show2()
        packet[IP].chksum = 0      # Set incorrect checksum

        with patch('builtins.print') as mock_print:
            checksum_verification(packet)
            mock_print.assert_called_with(f"Failed checksum verification from {packet[IP].src}")

    def test_payload_pattern_matching(self):
        packet = (IP(src="192.168.1.1", dst="192.168.1.2") /
                  TCP(dport=80, sport=12345) /
                  Ether() /
                  Raw(load="DROP TABLE"))

        with patch('builtins.print') as mock_print:
            payload_pattern_matching(packet)
            mock_print.assert_called_with(f"Possible Payload matching from {packet[IP].src}")

    def test_content_length_missmatch(self):
        payload = "GET / HTTP/1.1\r\n" "Host: example.com\r\n" "Content-Length: 10\r\n\r\n" "Hello, World!"
        packet = (Ether() /
                  IP(src="192.168.1.100", dst="192.168.1.1") /
                  TCP(dport=80, sport=12345) /
                  Raw(load=payload))

        with patch('builtins.print') as mock_print:
            content_length_mismatch(packet)
            mock_print.assert_called_with(f"Content length mismatch from {packet[IP].src}")

    def test_malformed_packets(self):
        oversized_packet = IP(src = "192.168.1.1", dst = "192.168.1.2") / ("X" * 2000)
        short_packet = IP(src = "192.168.1.9", dst = "192.168.1.2") / TCP(dport=80, sport=12345)

        with patch('builtins.print') as mock_print:
            malformed_check(oversized_packet)
            mock_print.assert_called_with(f"Malformed over-sized packet detected: {oversized_packet[IP].src}")
            malformed_check(short_packet)
            mock_print.assert_called_with(f"Malformed packet with short TCP detected: {short_packet[IP].src}")

    def test_protocol_fragmentation_check(self):
        fragment_1 = IP(src="192.168.1.1", dst="192.168.1.2", flags = "MF", frag = 0, len = 8)
        fragment_2 = IP(src="192.168.1.1", dst="192.168.1.2", flags = "MF", frag = 1, len = 8)
        fragment_3 = IP(src="192.168.1.1", dst="192.168.1.2", flags = "MF", frag = 2, len = 8)
        fragment_4 = IP(src="192.168.1.1", dst="192.168.1.2", frag = 3, len = 8) # Last fragment

        with patch('builtins.print') as mock_print:
            protocol_fragmentation_check(fragment_1)
            protocol_fragmentation_check(fragment_2)
            protocol_fragmentation_check(fragment_3)
            protocol_fragmentation_check(fragment_4)
            mock_print.assert_not_called()

        fragment_5 = IP(src="192.168.1.1", dst="192.168.1.2", flags="MF", frag=2, len=8)
        fragment_6 = IP(src="192.168.1.1", dst="192.168.1.2", frag=3, len=8) # Last fragment

        with patch('builtins.print') as mock_print:
            protocol_fragmentation_check(fragment_5)
            protocol_fragmentation_check(fragment_6)
            mock_print.assert_called_with(f"Possible Protocol Fragmentation from {fragment_5[IP].src}")

if __name__ == '__main__':
    unittest.main()
